import os, sys, re, random
import pandas as pd
from dataclasses import dataclass
import itertools
from functools import lru_cache, partial
from tqdm.auto import tqdm
from typing import Any
from anytree import Node, NodeMixin, LightNodeMixin, RenderTree
import psutil
from tqdm.auto import tqdm
import copy
from collections import defaultdict
import numpy as np
from easydict import EasyDict as edict
from functools import wraps


class FlatList(list):
    def __iadd__(self, other):
        if isinstance(other, list):
            self.extend(other)
        else:
            self.append(other)
        return self

    def __add__(self, other):
        result = FlatList(self)
        if isinstance(other, list):
            result.extend(other)
        else:
            result.append(other)
        return result


def Constraint(constraint_str):
    def generated_function(x):
        conditions = constraint_str.split(',')
        for cond in conditions:
            i, j = map(int, cond.split('∉'))
            if x[i].render('eng') in x[j].render('eng'):
                return False
        return True
    return generated_function

def apply_to_all_args(f):
    def decorator(func):
        def wrapper(*args, **kwargs):
            new_args = [f(arg) for arg in args]
            return func(*new_args, **kwargs)
        return wrapper
    return decorator

def Substitution(template,lang=None):
    def replace_template(template, a):
        # make numbers formattable 0 -> {0}
        wrap =  lambda s: (re.sub(r'(\d+)', r'{\1}', s) if type(s)==str else s)
        inner_replaced= re.sub(r"(\d+)\[\?←(.+?)\]",
            lambda m: a[int(m.group(1))].replace("?", (m.group(2))),
            template)
        output=wrap(inner_replaced).format(*a)
        return output

    """i[?←expr] replaces ? (but not \?) in slot i with expr"""
    @apply_to_all_args(lambda x:x.render(lang) if type(x)!=str else x)
    def sub(*a,**ka):
        y = replace_template(template,a)
        return y
    return sub



def Substitution(template, lang=None):
    def replace_template(template, a):
        # Make numbers formattable 0 -> {0}
        wrap = lambda s: (re.sub(r'(\d+)', r'{\1}', s) if isinstance(s, str) else s)

        # Function to safely replace only unescaped '?'
        def replace_match(m):
            slot_idx = int(m.group(1))
            replacement = m.group(2)
            return re.sub(r'(?<!\\)\?', replacement, a[slot_idx])  # Only replace unescaped '?'

        inner_replaced = re.sub(r"(\d+)\[\?←(.+?)\]", replace_match, template)
        output = wrap(inner_replaced).format(*a)

        # Convert escaped "\?" back to "?" after processing
        return output.replace(r'\?', '?')

    """i[?←expr] replaces ? (but not \?) in slot i with expr"""
    @apply_to_all_args(lambda x: x.render(lang) if not isinstance(x, str) else x)
    def sub(*a, **ka):
        return replace_template(template, a)

    return sub


# Pre-compile regex patterns
NUMBER_PATTERN = re.compile(r'(\d+)')
SUBSTITUTION_PATTERN = re.compile(r"(\d+)\[\?←(.+?)\]")


default_preprocess_template = lambda s: (re.sub(r'(\d+)', r'{\1}', s) if type(s)==str and '←' not in s else s)

def init_grammar(langs, name=''):
    class Rule:
        _instances = []
    
        @classmethod
        def init(cls, langs, name='', preprocess_template=default_preprocess_template):
            Rule.langs = langs
            Rule._instances = []
            Rule.preprocess_template = preprocess_template
            return langs
    
        def __init__(self, signature, *args, constraint=[], state_constraint=[], vars=dict(), weight=1):
            self.signature=signature
            self.name, self.args = self.parse_signature(signature)
            self.constraint = FlatList() + constraint
            self.state_constraint = FlatList() + state_constraint
            self.templates = {}
            self.weight=weight
            self.state = vars
            for lang, template in zip(Rule.langs, args*2):
                self.templates[lang] = Rule.preprocess_template(template)
            self.index = len(Rule._instances)
            Rule._instances.append(self)
    
        #def __getattr__(self,lang):
        #    return self.templates[lang]
    
        def parse_signature(self, s):
            name, args = re.match(r'(\w+)(?:\((.*?)\))?', s).groups()
            args = args.split(',') if args else []
            args = [a.strip() for a in args]
            return name, args
    
        @classmethod
        def get_rules(cls, name, sort=False, terminals=False,shuffle=False):
            instances = [x for x in cls._instances if x.name == name]
            if terminals:
                instances = [x for x in instances if not x.args]
            if shuffle:
                random.shuffle(instances)
            return instances
    
        @classmethod
        def start(cls):
            starts = cls.get_rules('start')
            if starts:
                return starts[0]
            else:
                return cls._instances[0]
    
        def __repr__(self):
            return f"RULE:{self.name}{self.args}"

    R = Rule
    R.init(langs,name)
    return R

Rule = init_grammar(None)


class Production(NodeMixin):
    def __init__(self, rule=None,type=None,state=dict()):
        self.rule = rule or ''
        self.type= type or self.rule.name
        self.state = {"parents": [],**state}
        if rule: 
            self.children = [Production(type=element) for element in self.rule.args]
            self.state = {**self.rule.state}
        self.cache=dict()

    def __setitem__(self, key, value):
        setattr(self,key,value)

    def distinct_constraint(self,x):
        self.pairs = itertools.combinations([idx for idx, a in enumerate(self.children) if self.children.count(a) > 1], 2)
        result= all(x[i] != x[j] for i, j in self.pairs)
        return result

    def __eq__(self,other):
        lang=self.rule.langs[0]
        return self@lang == other@lang

    def __getitem__(self,key):
        return self.children[key]

    def __len__(self):
        return len(self.children)

    def __deepcopy__(self, memo):
        new_obj = Production.__new__(Production)
        memo[id(self)] = new_obj
        new_obj.rule = self.rule
        new_obj.type = self.type
        new_obj.state = copy.deepcopy(self.state)
        new_obj.parent = None
        new_obj.cache = self.cache
        new_obj.children = [copy.deepcopy(child, memo) for child in self.children]
        return new_obj

    def check(self,mode='args'):
        if mode=='args':
            arguments=self.children
            if all(x.rule for x in arguments):
                return all(constraint(arguments) for constraint in self.rule.constraint+self.distinct_constraint)
            else:
                return True
        if mode=='state':
            return all(constraint(x) for x in [self,*self.ancestors] for constraint in x.rule.state_constraint)
            #constraints = [x.rule.state_constraint for x in [self,*self.ancestors]]
            #return all([constraint(self) for constraint in constraints])

    def render(self, lang=None):
        if lang==None:
            return str(RenderTree(self))
        if lang in self.cache:
            return self.cache[lang]
        try:
            template = self.rule.templates[lang]
        except:
            return "#"+self.type
        args = self.children
        if isinstance(template, str):
            args = [x.render(lang) for x in args]
            if '?←' in template:
                template = Substitution(template,lang)
            else:
                template_str = template
                template = template.format
        out = template(*args)
        if "#" not in out:
            self.cache[lang]=out
        return out
        
    __matmul__ = render


    def dict(self,use_cls=False):
        return edict({l:self@l for l in self.rule.langs}|(dict(cls=self) if use_cls else dict()))

    def __repr__(self):
        return f"PROD:{self.type}"+ (str(self.rule.args) if self.rule else '')


def safe_choices(sequence, weights=None, k=1):
    if not sequence:
        return sequence
    if weights is not None:
        weights = np.array(weights).flatten() / np.sum(weights)
        k = np.count_nonzero(weights)
    k = min(len(sequence), k)
    if weights is None or len(set(weights))==1:
        return random.sample(sequence, k)
    return np.random.choice(sequence, size=k, replace=False, p=weights)
    
def save(production,stack,step):
    ckpt=copy.deepcopy(production)
    ckpt.step=step
    ckpt.save=1
    ckpt_indices=[i for (i,x) in enumerate(stack) if x.save]
    #stack.insert(random.choice(ckpt_indices or [0]), ckpt)
    stack.insert(([0]+ckpt_indices)[-1], ckpt)

def generate_sequential(start,k=8,depth=14, skip_check=False,max_steps=1200,max_concentration_rate=0.8):
    start=Production(start)
    start.step=step=0
    start.save=0
    stack = [start]
    while stack:
        step+=1
        if step>max_steps:
            return []
        production = stack.pop()
        leaves=production.leaves
        if all(lv.rule for lv in leaves):
            if skip_check or all(x.check(mode='args') for x in (production,)+production.descendants):
                production.step=step
                return [production]
            continue
        lv = [lv for lv in leaves if not lv.rule][0]
        
        rules = start.rule.get_rules(lv.type,terminals=(lv.depth==depth),shuffle=False)
        if not rules:
            step+=5
            continue
        weights = [r.weight for r in rules]
        rules = safe_choices(rules,k=k, weights=weights)
        if len(rules)>1 and random.random()<(1/8):
            save(production,stack,step)
        for rule in rules:
            step+=0.25
            if step>max_steps:
                return []
            lv.rule=rule
            state={**rule.state,**lv.state}
            lv.children = [Production(type=c,state=state) for c in rule.args]
            if not lv.check(mode='state') and not skip_check:
                continue
            if all(x.rule for x in lv.siblings) and not lv.parent.check(mode='args') and not skip_check:
                continue #backtrack, chose another rule
            stack.append(production)
            break
    return []

def view(a,*la,**kwa):
    verbose=False
    if verbose:
        if type(a)==Production:
            try:
                if verbose>2:
                    clear_output()
                    print(a@eng,*la,**kwa)
                    print()
            except:
                pass
        else:
            print(a,*la,**kwa)



def generate_recursive(start, depth=14, max_steps=1200):
    start_prod = Production(start)
    
    def fill_tree(node, depth_budget=depth):
        empty_leaves = [lv for lv in node.leaves if not lv.rule]
        if not empty_leaves: return start_prod
        
        for lv in empty_leaves:
            # Use terminals when depth budget is exhausted
            use_terminals = (depth_budget <= 0)
            rules = node.rule.get_rules(lv.type, terminals=use_terminals, shuffle=True)
            if not rules: return None
            
            weights = [r.weight for r in rules]
            chosen_rule = random.choices(rules, weights=weights, k=1)[0]
            
            lv.rule = chosen_rule
            # Pass the reduced depth budget to children
            lv.children = [Production(type=c, state={**chosen_rule.state, **lv.state}) 
                          for c in chosen_rule.args]
        
        # Reduce depth budget for the next recursive call
        return fill_tree(node, depth_budget - 1)
    
    result = fill_tree(start_prod, depth)
    return [result] if result else []

def generate(start, n_iter=10_000, mode='recursive', *args, **kwargs):
    if type(start)==type:
        start=start.start()
        
    """Generate one production using specified mode."""
    for _ in range(n_iter):
        result = (generate_recursive if mode == 'recursive' else generate_sequential)(start, *args, **kwargs)
        if result: return result[0]
    raise ValueError('Incomplete generation')



def G0():
    """Toy grammar for illustration"""
    R = init_grammar(['tptp','eng'])
    def render_branch(x):
        #print(x.root, x.root.leaves, x.root.descendants, x.leftsbling)
        """equivalent to '0', illustrate function-based rendering"""
        return x@"eng"

    R('start(branch)', render_branch)
    R('branch(formula)', 'A:0')
    R('branch(formula)', 'B:0')
    R('branch(formula)', 'C:0')
    R('formula(formula)', '~(0)', 'It is not the case that "0"')
    R('formula(atom,op,atom)', '0 1 2', '0 1 2')
    R('formula(atom)', '0')
    R('atom', 'p', 'it is daytime')
    R('atom', 'q', 'it is raining')
    R('op', '&', 'and')
    R('op', '|', 'or')
    return R


def LogicNLI():
    ADJECTIVES = ['rich', 'quiet', 'old', 'tall', 'kind', 'brave', 'wise',
                  'happy', 'strong', 'curious', 'patient', 'funny', 'generous', 'humble']
    # (We selected adjectives with no clear semantic interference)
    NAMES = ['mary', 'paul', 'fred', 'alice', 'john', 'susan', 'lucy']

    R = init_grammar(['tptp','eng'])
    R('start(' + ','.join(['rule']*16) + ',' + ','.join(['fact']*8) + ')',
      '&\n'.join([f'({i})' for i in range(24)]),
      '\n'.join([f'{i}' for i in range(24)]))

    R('hypothesis(person,a)', '1(0)', '0 is 1')
    for a in ADJECTIVES:
        R('adj', a)
        R('adj', f'~{a}', f'not {a}', weight=0.2)

    R('property(adj,adj)', '(0(?)&1(?))', 'both 0 and 1')
    R('property(adj,adj)', '(0(?)|1(?))', '0 or 1')
    R('property(adj,adj)', '(0(?)<~>1(?))', 'either 0 or 1', weight=0.5)
    R('property(adj)', '0(?)', '0')

    R('rule(property,property)', '![X]:(0[?←X]=>1[?←X])',
      'everyone who is 0 is 1')
    R('rule(property,property)', '![X]:(0[?←X]<=>1[?←X])',
      'everyone who is 0 is 1 and vice versa')

    for p in NAMES:
        R('person', p)

    R('fact(person,property)', '1[?←0]', '0 is 1')
    R('fact(property)', '?[X]:(0[?←X])', 'someone is 0', weight=0.2)
    R('rule(fact,fact)', '(0)=>(1)', 'if 0 then 1')
    R('rule(fact,fact)', '(0)<=>(1)', 'if 0 then 1 and vice versa')
    return R


def test_grammar(grammar=G0()):
    "Illustrate the generation process"
    production = generate(grammar)
    return production, production@"eng"




from unigram import Rule, Substitution, Constraint, Production
import random
from xpflow import FlatList
from unigram.solver_utils.tptp import split_clauses, run, to_tptp, extract_inferences_and_formulas
import subprocess, copy
import itertools, re, random
import funcy as fc
from itertools import combinations
from collections import defaultdict
import numpy as np
import pandas as pd
import anytree

R,S,C=Rule,Substitution,Constraint
    
def pluralize(s):
    s=s.replace('does','do')
    s=s.replace('is a','are')
    s=s.replace('is ','are ')
    s=s.replace('es ','e ')
    s=s.replace('person','persons')
    s=s.replace('european','europeans')
    s=s.replace('tourist','tourists')
    return s


def polysyllogism(x):
    # P1-P1 P2-P2 P3-P3 P4|P4
    if x.parent is not x.parent.parent.children[0]: # lhs
        return # must be lhs
    if x.parent.parent.type!="cproperty":
        return 
    for a in x.ancestors:
        if a.type=='term':
            ls = leftsibling(a)
            if not ls:
                return
            cp=[d for d in ls.descendants if d.type=='cproperty']
            if not cp:
                return 
            return cp[-1][-1]


def get_concepts(text):
    return " ".join(tuple(re.findall(r'[a-z]+', text)))
    
R=Rule
lang=eng,tptp=['eng','tptp']
R.init(['tptp','eng'], "fof") #R(signature, fof_tptp, pseudo-english)

R('start(setup)', '0')# (x@l).replace('¿','?') for l in lang])


def B(*args, **kwargs):
    r = Rule("background",*args,**kwargs)
    r.background=True
    return r

x_y = lambda x:f"![X,Y]:(X!=Y=>({x}))"
x_y_z = lambda x: f"![X,Y,Z]:((X!=Y & Y!=Z & X!=Z)=>({x}))"


B(x_y("like(X,Y)=>~hate(X,Y)"))
B(x_y("hate(X,Y)=>~like(X,Y)"))
B("![X]:(european(X)=>person(X))")
B("![X]:(tourist(X)=>person(X))")

B("![X]:(anywhere(X))")

B(x_y("sibling(X,Y)<=>sibling(Y,X)"))
R('predicate','sibling','is a sibling of')

# neither symmetric nor transitive
R('predicate','client','is a client of')


adjs=['rich','quiet','old','tall','kind','brave','wise','happy','creative',
      'strong','curious','patient','funny','generous','calm','humble']

for a in adjs[:4]:
    
    B(x_y_z(f"({a}er(X,Y)&{a}er(Y,Z))=>{a}er(X,Z)")) #transitivity
    B(x_y(f"({a}er(X,Y)=>~{a}er(Y,X))")) #asymmetry
    B(x_y(f"({a}er(X,Y)&{a}(X))=>{a}(Y)")) #instanciation
    B(x_y(f"(~{a}(X)&{a}(Y))=>{a}er(X,Y)")) #reverse instanciation

    R('adj_comp',a)


R('predicate(adj_comp)','0er','is 0er than',weight=2)


R('term(property_toward_inside,outside_person)', '0[?←1]','1 0',constraint=C('1∉0,0∉1'))
R('property_toward_inside(predicate,person)', '0(1,?)','0 1')
R('property(predicate,outside_person)', '0(1,?)','0 1',weight=3)


def render_setup(x):
    persons=x.state['persons']
    room=x.state['room']
    persons_str = ', '.join([p.title() for p in persons])
    eng_setup=f"{persons_str} are the only persons in the {room}.\n{x[0]@eng}"
    if len(persons)==1:
        eng_setup=eng_setup.replace(' are ',' is ').replace('persons','person')
    in_room="&".join([f"room({x})" for x in persons])
    disj="|".join([f"X='{x}'" for x in persons])
    tptp_setup=in_room + '&' + f"(![X]:(room(X)=>({disj})))&\n{x[0]@tptp}"
    return tptp_setup,eng_setup

room = 'room'
names = 'mary', 'paul', 'fred', 'alice', 'john', 'susan', 'lucy'
psums = [list(names[:i]) for i in range(1, 5 + 1)]

for persons in psums:
    R('setup(valid_block)', lambda x:render_setup(x)[0], lambda x:render_setup(x)[1],
        vars=dict(persons=persons,room=room))

def render_setup(x):
    return f"(there_is_a_room)&\n{x[0]@tptp}", f"there is a room.\n{x[0]@eng}"
for persons in [names[:4],[],[]]:
    R('setup(valid_block)', lambda x:rs(x)[0], lambda x:render_setup(x)[1],
        vars=dict(persons=[],room=room))
    
def check_state(x):
    parents_=" ".join([a.signature for a in x.parent])
    pattern = r"\([^)]*term[^)]*\)"
    term_depth = len(re.findall(pattern, parents_))
    return term_depth<9
 
def no_free_var(x):
    s=x[0]@tptp
    return 'X' not in s or set('?!')&set(s)
    
R('valid_block(block)','0', constraint=no_free_var)

def neg_constraint(x):
    for d in x.descendants:
        if type(d.rule)==str:
            continue
        tptp=str(d.rule.templates['tptp'])
        if "=>" in tptp or "<=" in tptp:
            return False
        eng=str(d.rule.templates['eng'])
        if 'who' in eng:
            return False
    return True

nesting_limit = lambda x: x[0].render('eng').count('it is ') <= 2
R("term(term)", "~(0)", "it is not the case that “0”", constraint=nesting_limit,
  state_constraint=neg_constraint,weight=1)

R("term(term)", "0", "it is true that “0”",constraint=nesting_limit,weight=0.25)
R("term(term)", "(mary_dream=>(0))", "Mary dreamt that “0”",constraint=nesting_limit,weight=0.15)

def render_coref(coref,debug=""):
    c_tp = fc.rcompose(coref, lambda x: x@tptp if x else None)
    c_en = fc.rcompose(coref, lambda x: debug+x@eng  if x else None)
    return c_tp, c_en

class SeedContext:
    def __init__(self, seed):
        self.seed = seed

    def __enter__(self):
        random.seed(self.seed)

    def __exit__(self, exc_type, exc_value, traceback):
        random.seed(None)  # Reset to a random state

R('state','')
R("block(term)", "0", "0",weight=2)

def block_cond(x):
    for d in x.descendants:
        if type(d.rule)!=str and d.rule.name in ['cproperty','term']:
            if '=' in str(d.rule.templates['tptp']):
                return False
            if 'who' in str(d.rule.templates['eng']):
                return False
    return True


R('nterm(term)','0', state_constraint=block_cond)

R("cond(nterm,term)", "(0)=>(1)", "if “0” then “1”", weight=4)
R("cond(nterm,nterm,term)", "((0)=>(1))&\n((1)=>(2))", "if “0” then “1”\nif “1” then “2”")
R("cond(nterm,term)", "(0)=>(1)", "“1” if “0”")
R("cond(nterm,nterm)", "(0)<=>(1)", "“1” if “0” and vice versa")
R("cond(nterm,term)", "(0)=>(1)", "“0” only if “1”")
R("cond(nterm,nterm)", "(0)<~>(1)", "either “0” or “1” but not both")
R("cond(term,nterm)", "~(1)=>(0)", "“0” unless “1”")
R('cond(nterm,term,term)', '((0)=>(1))&((~(0))=>(2))', 'if “0” then “1” otherwise “2”')
#R("block(term,term)", "(0)&\n(1)", "0\n1")

conj_tptp = lambda i: '&\n'.join([f'({i})' for i in range(i)])
conj_eng  = lambda i: '\n'.join([f'{i}' for i in range(i)])
conj = lambda i: (conj_tptp(i), conj_eng(i))

for n in list(range(3, 32)): 
    for nconds in range(int(n//2)):
        nterms = n - nconds  
        arg1 = "block(" + "cond,"*nconds + ("term,"*(nterms-1)).rstrip(',') + ")"
        R(arg1, *conj(n-1),weight=w)

for n in names:
    R("all_persons",n,n.title())
B('&'.join('({}!={})'.format(*a) for a in itertools.product(names,names) if len(set(a))!=1))

def inside_person(x):
    return x[0]@tptp in x[0].state['persons']
R('person(all_persons)','0','0',constraint=inside_person)

def all_unique(x):
    seen = set()
    return not any(i in seen or seen.add(i) for i in x)

def block_constraint(x):
    filled = [x for x in x.children if x.rule]
    return all_unique(x@eng for x in filled if "#" not in x@eng)

for i in 'abcdefghijklmnopqrstuvwxyz'
    R('atomic_predicate',f"pred_{i}")
R('property(atomic_predicate)', '0(?)','0',weight=6)
R('property(atomic_predicate)', '~0(?)','~0',weight=1)

R('term(person,property)', '1[?←0]',
    fc.rcompose(S('0 1', lang='eng'), lambda x: x.replace('they', 'he/she')),
    constraint=[C('0∉1,1∉0'), lambda x: 'who' not in x[1]@eng], weight=8)

R('term(person,property,property)','(1[?←0])&(2[?←0])','0 who 1 2',weight=1)

R('term(property,person,person)', '0[?←1]&0[?←2]', (S('1 and 2 0',lang='eng')|pluralize),
    constraint=[
        C('0∉1,1∉2,0∉2'),
        lambda x:'pred' not in x[0]@tptp,
        lambda x:'who' not in x[0]@eng],
  weight=2)

R('quantifier', '!', 'everyone',weight=10), R('quantifier','~!','not everyone')
R('quantifier','~?','no one') 
R('group','room','in the room',weight=8), R('group','~room','outside the room')
R('group','anywhere','anywhere')

R('X_quantifier(quantifier,group)','0[X]:(1(X)=>(?))','0 1')

def block_property(x):
    for d in x.descendants:
        if d.rule and 'property' in d.rule.args:
            return False
    return True

kw=dict(weight=0.5,state_constraint=block_property)
R('property(property,property)', '((0)<~>(1))', 'either 0 or 1 but not both',**kw)
R('property(property,property)', '((0)|(1))', '0 or 1 or both',**kw)
R('property(property,property)', '((0)&(1))', '0 and 1',**kw)
R('property(property,property)', '~((0)|(1))', 'neither 0 nor 1',**kw)
R('property(property,property,property)', '((0)&(1)&(2))', '0, 1 and 2',**kw)
R('property(property,property,property)', '((0)|(1)|(2))', '0, 1 or 2',**kw)

R('property(state)',*render_coref(polysyllogism,debug=""),
  state_constraint=lambda x:bool(x@tptp), weight=14)

R('cproperty(property,property)', '((0)=>(1))', 'who 0 1',weight=4)
R('cproperty(property,property)', '((0)=>(1))', '1 if they 0',weight=4)
R('cproperty(property,property)', '((1)<=(0))', '0 only if they 1')
R('cproperty(property,property)', '((0)<=>(1))', '0 if they 1 and vice versa')
R('cproperty(property,property)', '((0)<=>(1))', '0 if and only if they 1')


R('X_property(cproperty)','0[?←X]','0')
R('X_property(property)','0[?←X]','0',weight=0.1)

R('term(X_quantifier,X_property)', '0[?←1]', '0 1', weight=12) 

R('E_quantifier(group)','?[X]:(0(X)&(?))','someone 0')
R('E_property(property)','0[?←X]','0')
R('term(E_quantifier,E_property)', '0[?←1]', '0 1') 

R('term(property,group)', '((?[X]:(1(X)&0[?←X])))', 'at least one person 1 0')
R('term(property,group)', '((?[X]:(1(X)&0[?←X]))&(![X,Y]:((1(X)&1(Y)&(0[?←X])&(0[?←Y]))=>(X=Y))))', 'only one person 1 0')
R('term(property,group)', '(?[X,Y]:(1(X)&1(Y)&(0[?←X]&0[?←Y])&(X!=Y)))', 'more than one person 1 0')

R('term(person,adj_comp,group)', '![X]:((2(X)&(X!=0))=>1er(X,0))', '0 is the 1est 2')
R('term(person,adj_comp,group)', '![X]:((2(X)&(X!=0))=>1er(0,X))', '0 is the least 1 2');

background="&\n".join([f"({x.templates['tptp']})" for x in R.get_rules('background')])
background=split_clauses(background,name_prefix='b')
